import json
import yaml
from arms.yapi.util import prefix, extract_include, \
    big_mdtitle_group, small_mdtitle_group, mdyaml_group, pathparam_names, merge_dicts, \
    parse_comment, is_comment_format


class Project:
    """
    includes:
      - str
    list:  //projectItems
      - name: str e.g.学校管理
        list:  //projectSubItems
          - name: str e.g.查询学校详情
            method: str
            path: str
            tag: str
            input:
              - Dict
            output: Dict

    """
    @staticmethod
    def subitem_of(yaml_dict):
        """
        后续会做yaml_dict内的类型扩展
        """
        return yaml_dict

    @staticmethod
    def loads(text):
        """
        >>> text = open("example/api.md", encoding='utf8').read()
        >>> assert Project.loads(text)
        >>> assert json.dumps(Project.loads(text), indent=2)
        """
        project = {'includes': [], 'list': []}
        lines = text.splitlines()
        project['includes'] = list(prefix(lines, extract_include))
        for big_lines in big_mdtitle_group(lines):
            title = big_lines[0].replace('#', ' ').strip()
            project_subitems = []
            for small_lines in small_mdtitle_group(big_lines[1:]):
                subtitle = small_lines[0].replace('#', ' ').strip()
                yaml_groups = list(mdyaml_group(small_lines[1:]))
                assert yaml_groups, "Missing yaml part: " + subtitle
                fulldoc = '\n'.join(yaml_groups[0][1:-1])
                yaml_dict = yaml.load(fulldoc)
                yaml_dict['fulldoc'] = fulldoc
                yaml_dict['name'] = subtitle
                yaml_dict.setdefault('tag', '')
                assert 'method' in yaml_dict, subtitle
                assert 'path' in yaml_dict, subtitle
                assert 'input' in yaml_dict and isinstance(yaml_dict['input'], list), subtitle
                assert 'output' in yaml_dict and isinstance(yaml_dict['output'], dict), subtitle
                project_subitems.append(Project.subitem_of(yaml_dict))
            project['list'].append({'name': title, 'list': project_subitems})
        return project


class YapiProject:
    """
    //yapiProjectItems
    - name: str e.g.学校管理
      desc: null
      index: 0
      list: //yapiProjectSubItems
        - method: str e.g."POST"
          title: str e.g."增加或修改学校"
          markdown: str e.g.接口详细描述
          path: str e.g."/school/save"
          req_headers: [{ "value": "", "name": "token", "required": "1"}]
          req_params:
            - name: str e.g."schoolId"
              desc: str e.g."学校ID"
              example: str e.g."北京大学"
          req_query:
            - name: str e.g."schoolName"
              desc: str e.g."学校名称 (搜索参数)"
              required: str e.g."0"
              example: str e.g."北京大学"
          req_body_is_json_schema: false
          req_body_type: form
          req_body_form:
            - name: str e.g."schoolName"
              desc: str e.g."学校名称"
              required: str e.g."0"
              example: str e.g."北京大学"
              type: text
          res_body_type: json
          res_body_is_json_schema: false
          res_body: str e.g."{...}"
          type: static //req_params为空时
          type: var    //req_params非空时
          __v: 0
          api_opened: false
          query_path: {"path": "/school/{schoolId}", "params": []},
          status: undone
    """
    @staticmethod
    def yapijson_of(project):
        yapi_project_items = []
        for project_item in project['list']:
            name, project_subitems = project_item['name'], project_item['list']
            yapi_project_subitems = [YapiProject.subitem_of(item) for item in project_subitems]
            yapi_project_item = {'name': name, 'desc': None, 'index': 0, 'list': yapi_project_subitems}
            yapi_project_items.append(yapi_project_item)
        return json.dumps(yapi_project_items, indent=2)

    @staticmethod
    def subitem_of(project_subitem):
        method = project_subitem['method'].upper()
        title = project_subitem['name']
        markdown = ''
        path = project_subitem['path']
        pathparams = pathparam_names(path)
        itemtype = 'var' if pathparams else 'static'
        inputreqs = merge_dicts(project_subitem['input'])
        req_params = []
        req_query = []
        req_body_form = []
        # 根据method, inputreqs和pathparams设置req_*
        use_body = (method == 'POST' or method == 'PUT')
        for key, value in inputreqs.items():
            if key == '//':
                continue
            elif key in pathparams:
                reqitem = YapiProject.reqitem_of(key, value, is_param=True)
                req_params.append(reqitem)
            elif use_body:
                reqitem = YapiProject.reqitem_of(key, value, is_form=True)
                req_body_form.append(reqitem)
            else:
                reqitem = YapiProject.reqitem_of(key, value)
                req_query.append(reqitem)
        # 根据output生成res_body
        res_body = YapiProject.resbody_of(project_subitem['output'])
        return {
            'method': method,
            'title': title,
            'markdown': markdown,
            'path': path,
            'req_headers': [{ "value": "", "name": "token", "required": "1"}],
            'req_params': req_params,
            'req_query': req_query,
            'req_body_is_json_schema': False,
            'req_body_type': 'form',
            'req_body_form': req_body_form,
            'res_body_type': 'json',
            'res_body_is_json_schema': False,
            'res_body': res_body,
            'type': itemtype,
            '__v': 0,
            'api_opened': False,
            'query_path': {"path": path, "params": []},
            'status': 'undone',
        }

    @staticmethod
    def reqitem_of(key, value, is_param=False, is_form=False):
        """
        {key='id?', value='int=22 //hehe'} => {name='id', desc='hehe', example='22', required='0', type='text'}
        """
        assert isinstance(key, str) and key, "invalid key: " + str(key)
        name, required = (key[:-1], "0") if key[-1] == '?' else (key, "1")
        pytype, realvalue, comment, _ = parse_comment(value)
        if realvalue.startswith('"'):
            realvalue = json.loads(realvalue)
        reqitem = {'name': name, 'desc': comment, 'example': realvalue}
        if not is_param: reqitem['required'] = required
        if is_form: reqitem['type'] = 'text'
        return reqitem

    @staticmethod
    def resbody_of(output_dict):
        """
        >>> output_dict = {'//': 'comment', 'aa': True, 'bb': 'str="true" //n', 'cc': 'long=22 //hehe'}
        >>> print(YapiProject.resbody_of(output_dict))
        {
          // comment
          "aa": true,
          "bb": "true",  //n
          "cc": 22  //hehe
        }
        """
        lines = json.dumps(output_dict, indent=2).splitlines()
        new_lines = []
        for line in lines:  # line=`"bb": "true",  //n`
            if line.startswith('#'):
                continue
            elif ':' in line and is_comment_format(line.split(':', 1)[-1]):
                tail = ''
                key, value = line.split(': ', 1)
                if value[-1] == ',':
                    value, tail = value[:-1], ','
                value = json.loads(value)
                pytype, realvalue, comment, _ = parse_comment(value)
                new_line = '%s: %s%s  //%s' % (key, realvalue, tail, comment)
                new_lines.append(new_line)
            elif line.strip().startswith('"//"'):
                key, value = line.split(': ', 1)
                if value[-1] == ',': value = value[:-1]
                new_lines.append('%s// %s' % (key[:-4], json.loads(value)))
            else:
                new_lines.append(line)

        return '\n'.join(new_lines)


class ClassDef:
    """
    className: str
    args:
      - name: str
        typehint: str
        desc: str
        example: str
    """
    pass


class EnumDef:
    """
    className: str
    args:
      - name: str
        value: str
        desc: str
    """
    pass


class ModelDefs:
    """
    enums: List[EnumDef]
    classes: List[ClassDef]
    """
    pass


import os
from arms.yapi.model import Project


key_tpl = '@{method}Mapping("{path}")'

code_tpl = """
/**
{fulldoc}
 */
//{name}
{key}
public Result {tag}() <(
    return null;
)>
"""


def load_outpath(outpath, in_keys):
    exit_code = 0
    filenames = [name for name in os.listdir(outpath) if name.endswith('.java')]
    for filename in filenames:
        text = open(os.path.join(outpath, filename)).read()
        for line in text.splitlines():
            line = line.strip()
            if line.startswith('@') and 'Mapping(' in line:
                if line not in in_keys:
                    print('Need deleting: %s  (%s)' % (line, filename))
                    exit_code = 1
                else:
                    del in_keys[line]
    return exit_code


def process(inpath, outpath):
    filenames = [name for name in os.listdir(inpath) if name.endswith('.md')]
    in_keys = {}
    for filename in filenames:
        text = open(os.path.join(inpath, filename)).read()
        project = Project.loads(text)
        for item in project['list']:
            for subitem in item['list']:
                assert subitem.get('tag'), 'tag(函数名称)不能为空: ' + subitem['name']
                in_key = key_tpl.format(method=subitem['method'].capitalize(), path=subitem['path'])
                code = code_tpl.format(
                    name=subitem['name'], key=in_key, tag=subitem['tag'], fulldoc=subitem['fulldoc'].strip(),
                ).replace('<(', '{').replace(')>', '}')
                in_keys[in_key] = code

    exit_code = load_outpath(outpath, in_keys)
    if in_keys:
        with open(os.path.join(outpath, 'OtherController.java'), 'w') as f:
            f.write('package controller;\n\nimport com.parsec.universal.utils.Result;\n'
                    'import org.springframework.web.bind.annotation.*;\n\n@RestController\n'
                    'public class OtherController {\n')
            for code in in_keys.values():
                f.write(code + '\n')
            f.write('}\n')

    exit(exit_code)




import yaml
import os
from arms.yapi.util import parse_comment, is_comment_format


def is_utf8mb4(stype, text):
    """
    >>> is_utf8mb4('VARCHAR(100)', "重庆")
    True
    >>> is_utf8mb4('JSON', "chongqing")
    False
    """
    return (stype.startswith('VARCHAR(') or stype in ['TEXT', 'JSON']) \
           and any(ord(c) >= 256 for c in text)


def is_strtype(stype):
    return


def sql_type_of(pytype):
    typename_map = {
        'int': 'INT(11)',
        'long': 'BIGINT(20)',
        'bool': 'TINYINT(1)',
        'float': 'DOUBLE',
        'str': 'VARCHAR(200)',
        'text': 'TEXT',
        'datetime': 'DATETIME',
        'date': 'DATE',
        'time': 'TIME',
        'json': 'JSON',
    }
    if pytype.startswith('+'):
        return '+' + sql_type_of(pytype[1:])
    if pytype.startswith('str('):
        return 'VARCHAR(' + pytype[4:]
    return typename_map.get(pytype, None)


normal_sline_tpl = "  `{name}` {stype} DEFAULT NULL COMMENT '{comment}',"
primary_sline_tpl = "  `{name}` {stype} NOT NULL AUTO_INCREMENT COMMENT '{comment}',"
utf8_sline_tpl = "  `{name}` {stype} CHARACTER SET utf8mb4 DEFAULT NULL COMMENT '{comment}',"

table_tpl = """
DROP TABLE IF EXISTS `{model_name}`;
CREATE TABLE `{model_name}` (
{scode}
) ENGINE=InnoDB DEFAULT CHARSET=ascii;
"""


def scode_of(model_name, model_dict):
    slines = []
    primary_key = ''
    for name, value in model_dict.items():
        if name == '//':
            continue
        assert is_comment_format(value), "invalid value: " + value
        pytype, realvalue, comment, _ = parse_comment(value)
        stype = sql_type_of(pytype)
        assert stype is not None, "invalid type: " + value
        if name.lower() == model_name.lower() + 'id':
            primary_key = name
            assert not stype.startswith('+'), "invalid private key: " + value
            sline = primary_sline_tpl.format(stype=stype, name=name, comment=comment)
        elif stype.startswith('+'):
            continue
        elif isinstance(realvalue, str) and is_utf8mb4(stype, realvalue):
            sline = utf8_sline_tpl.format(stype=stype, name=name, comment=comment)
        else:
            sline = normal_sline_tpl.format(stype=stype, name=name, comment=comment)

        slines.append(sline)

    slines.append("  PRIMARY KEY (`%s`)" % primary_key)
    scode = '\n'.join(slines)
    return dict(scode=scode, primary_key=primary_key)


def scontent_of(model_name, model_dict):
    scode_ret = scode_of(model_name, model_dict)
    scode = scode_ret['scode']
    primary_key = scode_ret['primary_key']
    if not primary_key:
        return ''
    return table_tpl.format(model_name='Tbl'+model_name, scode=scode)


def diff_scode(scode, model_data):
    exit_code = 0
    cur_modelname = ''
    new_keys = set()
    for model_name, model_dict in model_data.items():
        for key, value in model_dict.items():
            if key == '//': continue
            pytype = parse_comment(value)[0]
            if pytype.startswith('+'): continue
            new_keys.add('%s.%s' % (model_name, key))
    for line in scode.splitlines():
        line = line.strip()
        if line.startswith('CREATE TABLE '):
            cur_modelname = line.split('`Tbl', 1)[-1].split('`')[0]
        elif cur_modelname and line.startswith(') ENGINE'):
            cur_modelname = ''
        elif line.startswith('`'):
            old_key = cur_modelname + '.' + line.split('`', 1)[-1].split('`')[0]
            if old_key not in new_keys:
                print('Need deleting: %s' % old_key)
                exit_code = 1
            else:
                new_keys.remove(old_key)

    for key in new_keys:
        print('Need inserting: %s' % key)
        exit_code = 1

    return exit_code


def process(infile, outfile):
    scontents = []
    model_data = yaml.load(open(infile).read())
    if os.path.isfile(outfile):
        scode = open(outfile).read()
        exit_code = diff_scode(scode, model_data)
        exit(exit_code)
    else:
        for model_name, model_dict in model_data.items():
            scontents.append(scontent_of(model_name, model_dict))

        with open(outfile, 'w') as f:
            f.write('\n'.join(scontents))



import os
import lesscli
import hashlib
import requests
import yaml
from jinja2 import Template

from arms.utils import project_to_package, camelize

PY2 = (type(b'') == str)
if PY2:
    input = raw_input
    def makedir(real_path):
        os.system("mkdir -p " + real_path)

else:
    def makedir(real_path):
        from pathlib import Path
        Path(real_path).mkdir(parents=True, exist_ok=True)


def print_help():
    text = """
    
    arms init                   : setup CI/CD
    arms run [stage]            : run scripts of .gitlab-ci.yml
    arms docker prune           : prune docker container/image/volume
    arms paste [format]         : change coped content into another format
    arms yapi [json|controller
        |model|schema] --java 
        [--check] [inpath] 
        -o [outpath]            : build yapi from markdown
    arms -h                     : show help information
    arms -v                     : show version
    
    
    """
    print(text)


def print_version():
    from arms import __version__
    text = """
    arms version: {}
    
    """.format(__version__)
    print(text)


def grab(project, lang):
    package = project_to_package(project)
    get_port = lambda name: int(hashlib.sha1(name.encode()).hexdigest(), 16) % 22000 + 10000
    port = get_port(package)  # [10000, 32000)
    mysql_port = get_port(package+'_mysql')
    redis_port = get_port(package+'_redis')
    package_upper = camelize(package)
    package_lower = camelize(package, False)
    base = 'http://gitlab.parsec.com.cn/qorzj/chiji-tool/raw/master/templates/' + lang
    index_text = requests.get(base + '/.index.txt').text
    data = {
        'project': project, 'package': package,
        'package_upper': package_upper, 'package_lower': package_lower, 'image': package.lower(),
        'port': port, 'mysql_port': mysql_port, 'redis_port': redis_port,
    }
    for line in index_text.splitlines():
        if not line: continue
        url = base + '/' + line
        real_path = Template(line).render(**data).replace('_PLACEholder_', '')
        print(url + '\t' + real_path)
        if real_path[-1] == '/':
            makedir(real_path)
        else:
            req = requests.get(url)
            if req.status_code == 404:
                print('Template not found!')
                return
            file_text = req.text
            rended_text = Template(file_text).render(**data)
            if real_path == '.gitignore' and os.path.isfile(real_path):  # merge .gitignore
                old_lines = open(real_path).read().splitlines()
                for new_line in rended_text.splitlines():
                    if new_line not in old_lines:
                        old_lines.append(new_line)
                rended_text = '\n'.join(old_lines) + '\n'

            with open(real_path, 'wb') as f:
                f.write(rended_text.encode('utf-8'))
    print('---- Done ----')


def run_init():
    if not os.path.isdir('.git'):
        print('Please change workdir to top!')
        return
    front = input('Please input front or back: [front / back] ')
    lang_front = ['react', 'build']
    lang_back = ['java', 'node', 'python', 'nginx']
    if front == 'front':
        lang_short = input('Please input language: [%s] ' % ' / '.join(lang_front))
        assert lang_short in lang_front
    elif front == 'back':
        lang_short = input('Please input language: [%s] ' % ' / '.join(lang_back))
        assert lang_short in lang_back
    else:
        print('Error, please check!')
        return
    lang = front + '-' + lang_short
    project = os.path.abspath('').rsplit('/')[-1]
    grab(project, lang)


def run_run(stage):
    if not os.path.isfile('.gitlab-ci.yml'):
        print('.gitlab-ci.yml not found!')
        return
    try:
        obj = yaml.load(open('.gitlab-ci.yml').read())
    except:
        print('Cannot load .gitlab-ci.yml')
        return

    def _get_scripts(stage):
        for v in obj.values():
            if isinstance(v, dict) and v.get('stage') == stage:
                return v.get('script', [])

    scripts = _get_scripts(stage)
    if not scripts:
        print('Scripts of stage[{}] is empty'.format(stage))
        return
    for cmd in scripts:
        print('>>', cmd)
        errno = os.system(cmd)
        if errno:
            print('---- Failed! [errno: %d] ----' % errno)
            exit(errno)
    print('---- Done. ----')


def run_docker_prune():
    os.system('docker container prune -f')
    os.system('docker image prune -f')
    os.system('docker volume prune -f')


def run_paste(format):
    pyperclip = None
    try:
        import pyperclip
    except:
        print('---- Failed! Dependence pyperclip not installed ----')
        print('---- Need to run: pip install pyperclip ----')
        exit(1)
    format = format.lower()
    if format not in ['po', 'db', 'acl']:
        print('---- Failed! Only support format [po, db, acl] ----')
        exit(1)
    raw_content = pyperclip.paste()
    if not raw_content:
        print('---- Failed! Please copy content first')
        exit(1)

    # translate format: po
    if format == 'po':
        typedict = {'string': 'String', 'number': 'Integer', '': '____'}
        lines = []
        for raw_line in raw_content.splitlines():
            segs = raw_line.lstrip().split('\t')
            if len(segs) < 3:
                print('---- Failed! Unknown input format ----'); print(segs); print('')
                exit(1)
            varname, explain, vartype = segs[:3]
            javatype = 'Date' if varname.endswith('At') \
                else typedict.get(vartype, vartype)
            lines.append('\tprivate {T} {name};  // {expl}'.format(T=javatype, name=varname, expl=explain))
        out_content = '\n'.join(lines)
    elif format == 'db':
        typedict = {'int': 'Integer', 'smallint': 'Integer',
                    'varchar': 'String', 'text': 'String', 'json': 'String',
                    'datetime': 'Date'}
        lines = []
        for raw_line in raw_content.splitlines():
            if ' COMMENT ' not in raw_line: raw_line += ' COMMENT '
            left, explain = raw_line.split(' COMMENT ', 1)
            segs = left.lstrip().split()
            if len(segs) < 2 or segs[0][0] != '`':
                print('---- Failed! Unknown input format ----'); print(segs); print('')
                exit(1)
            varname, vartype = segs[:2]
            varname = varname.replace('`', '')
            vartype = vartype.split('(', 1)[0] if '(' in vartype else vartype
            javatype = typedict.get(vartype, vartype)
            explain = explain.strip()
            explain = explain[1:-2] if explain.startswith("'") and explain.endswith("',") else explain
            lines.append('\tprivate {T} {name};  // {expl}'.format(T=javatype, name=varname, expl=explain))
        out_content = '\n'.join(lines)
    elif format == 'acl':
        fields = {}
        lines = []
        try:
            for raw_line in raw_content.splitlines():
                key, content = raw_line.strip().split(None, 1)
                fields[key] = content
            url = fields.get('请求Url', '')
            method = fields.get('请求类型', '')
            name = fields.get('接口名称', '')
            desc = fields.get('接口描述', '')
            pathvar = ''
            if url.count(':') == 1:
                pathvar = url.split(':')[-1]
                url = url.replace(':'+pathvar, '{'+pathvar+'}')

            lines.append("// (NULL,'{url}','{method}','',NULL,'{name}'),".format(url=url, method=method, name=name))
            if name: lines.append("// " + name)
            if desc: lines.append("// " + desc)
            lines.append('@{method}Mapping("{url}")'.format(method=method.capitalize(), url=url))
            if pathvar:
                lines.append('public Result ____(@PathVariable("%s") int %s) {' % (pathvar, pathvar))
            else:
                lines.append('public Result ____(____) {')
            lines.append('    return null;')
            lines.append('}')
        except:
            print("""复制文本的样例：
            接口名称 生成查询二维码
            请求类型 get
            请求Url  /wx/makeQRCode/:fileno
            接口描述 直接返回png格式的二维码图片
            """)
            exit(1)
        out_content = '    ' + '\n    '.join(lines) + '\n\n'
    else:
        raise NotImplementedError
    pyperclip.copy(out_content)
    print('\n%s\n' % out_content)
    print('---- Content above is copied ----')


def run_yapi(option, infile, outfile, checkonly):
    from arms.yapi.model import Project, YapiProject
    from arms.yapi.java_util import controller_tool, model_tool, schema_tool
    # option in 'json|controller|model|schema'
    if option == 'json':
        assert outfile.endswith('.json'), "outfile[%s] is not a .json file" % outfile
        project = Project.loads(open(infile).read())
        yapijson = YapiProject.yapijson_of(project)
        with open(outfile, 'w') as fout:
            fout.write(yapijson)
    elif option == 'controller':
        controller_tool.process(inpath=infile, outpath=outfile)
    elif option == 'model':
        model_tool.process(infile=infile, outpath=outfile, replace=not checkonly)
    elif option == 'schema':
        schema_tool.process(infile=infile, outfile=outfile)
    else:
        raise NotImplementedError
    print('done.')


def run(*a, **b):
    if 'h' in b or 'help' in b:
        print_help()
    elif 'v' in b or 'version' in b:
        print_version()
    elif tuple(a) == ('init',):
        run_init()
    elif len(a) == 2 and a[0] == 'run':
        stage = a[1]
        run_run(stage)
    elif tuple(a) == ('docker', 'prune'):
        run_docker_prune()
    elif len(a) >= 1 and a[0] == 'paste':
        run_paste(a[1] if len(a) > 1 else '')
    elif len(a) == 3 and a[0] == 'yapi':
        run_yapi(option=a[1], infile=a[2], outfile=b['o'], checkonly='check' in b)
    else:
        print_help()


def entrypoint():
    lesscli.run(callback=run, single='hv')